'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var lodash = require('lodash');
var jQuery = require('jquery');
var md5 = require('js-md5');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var jQuery__default = /*#__PURE__*/_interopDefaultLegacy(jQuery);
var md5__default = /*#__PURE__*/_interopDefaultLegacy(md5);

function getAttrs(xml) {
  let attributes = {};
  [...xml.attributes].forEach(item => {
    attributes[item.name] = item.value.split(' ').filter(item => lodash.trim(item)).map(item => lodash.trim(item));
  });
  return attributes;
}

class PropsWatcher {
  constructor() {
    /**
     * {
     *  0: {
     *    myclick: [fn1, fn2, fn3]
     *  }
     * }
     */
    this.listenMap = {};
  }
  /**
   * 
   * @param {*} id 组件id
   * @param {*} key 事件名称
   * @param {*} fn 回调函数
   */
  on(compId, key, fn) {
    let compListener = this.listenMap[compId];
    // 事件名称必须是字符串，事件监听回调必须是函数
    if (lodash.isString(key) && lodash.isFunction(fn)) {
      if (!lodash.isObject(compListener)) {
        this.listenMap[compId] = {};
      }
      if (!lodash.isArray(this.listenMap[compId][key])) {
        this.listenMap[compId][key] = [];
      }
      this.listenMap[compId][key].push(fn);
    }
  }

  /**
   * 事件触发
   * @param {*} compId 组件id
   * @param {*} key 事件名称
   * @param {*} data 返回值
   */
  emit(compId, key, data) {
    let compListener = this.listenMap[compId];
    if (compId == 'all') {
      Object.keys(this.listenMap).forEach((cid) => {
        this.emit(cid, key, data);
      });
      return;
    } 
    if (lodash.isObject(compListener) && lodash.isArray(compListener[key]) && compListener[key].length > 0) {
      compListener[key].forEach(fn => {
        fn(data);
      });
    }
  }
}

const COMPONENT_ID = 'component-id';

const COMPONENT_NAME = 'component-name';

const COMPONENT_ROUTER = 'component-router';

const createdID = () => {
  let id = 0;
  return () => {
    return id++;
  }
};
let JqComponent_ID = createdID();

function create$$($dom) {
  return function find(selector) {
    let childen = $dom.find(selector);
    $dom.parent().find(selector).each((index, $d) => {
      if ($dom.get(0) === $d) {
        childen.push($d);
      }
    });
    return childen;
  };
}

function setAttr($fragment, attrs) {
  Object.keys(attrs).forEach(key => {
    let values = attrs[key];
    values.forEach(v => {
      let lastedVal = $fragment.attr(key);
      let attrVals = lastedVal ? `${lastedVal} ${v}` : v;
      $fragment.attr(key, attrVals);
    });
  });
}

var mountedStylesTagNameMap = {};

/**
 * 这里用tagName判断，但是tagName会有hash碰撞的可能
 * @param {*} styles 
 * @param {*} tagName 
 */
function mountedStyles(styles, tagName) {
  if (!mountedStylesTagNameMap[tagName] && styles) {
    jQuery__default['default']("head").append(`<style>${styles}</style>`);
    mountedStylesTagNameMap[tagName] = true;
  }
}

function proxy$$($$) {
  var handler = {
    get(target, key) {
      if (!target[key]) {
        return jQuery__default['default'][key];
      }
      return target[key];
    }
  };
  return new Proxy($$, handler);
}



const setComponentInfo2Dom = (dom, tagName, id) => {
  if (id) {
    jQuery__default['default'](dom).attr(COMPONENT_ID, id);
  }
  if (tagName) {
    jQuery__default['default'](dom).attr(COMPONENT_NAME, tagName);
  }
};

// const eventHooks = function (id, dom, attrs) {
//   Object.keys(attrs).forEach(attr => {
//     if (startsWith(attr, 'on')) {
//       dom[attr] = () => {
//         let lisfn = attrs[attr][0];
//         var fnName = lisfn.replace(/\(.+?\)/g, '')
//         var arg = lisfn.match(/\((.+?)\)/g);
//         arg = arg[0].replace(/\(|\)/g, '').split(',').map(item => trim(item));
//         listenProps.emit(id, fnName, arg, dom);
//       }
//     }
//   })
// }

let allComponents = {};

let globalComponents = {};

const listenProps = new PropsWatcher();

const setGlobalComponent = function (name, jqc) {
  globalComponents[name] = jqc;
  return globalComponents;
};

const removeGlobalComponent = function (name) {
  delete globalComponents[name];
  return globalComponents;
};
class JqComponent {
  constructor(options = {}) {
    this.options = options;
    let { template, styles, components, ready } = this.options;
    this.template = template;
    this.$fragment = jQuery__default['default'](template);
    this.components = globalComponents;
    this.styles = styles;
    this.ready = ready;
    this.id = JqComponent_ID();
    this.slots = {};

    Object.keys(components || {}).forEach(key => {
      this.components[lodash.toUpper(key)] = components[key];
    });

    allComponents[this.id] = this;
  }

  render(root) {
    const tagName = root.tagName;
    if (tagName) {
      // 处理插槽
      if (tagName === 'SLOT') {
        const slotName = root.name ? root.name : 'default';
        const slotElem = this.slots[slotName];
        if (slotElem) {
          root.parentNode.replaceChild(slotElem, root);
        }
      }
      const options = this.components[tagName];
      const attrs = getAttrs(root);
   
      if (options) {
        const jqc = new JqComponent(options);

        // 保存插槽
        const slots = lodash.trim(root.innerHTML);
        if (slots) {
          jQuery__default['default'](slots).each((idx, node) => {
            if (node.nodeType == 1) {
              const slotName = node.slot ? node.slot : 'default';
              jqc.slots[slotName] = node;
            }
          });
        }

        setAttr(jqc.$fragment, attrs);

        root.parentNode.replaceChild(jqc.$fragment.get(0), root);
        setComponentInfo2Dom(jqc.$fragment, null, jqc.id);
        mountedStyles(jqc.styles, tagName);
        jqc.render(jqc.$fragment.get(0));

        if (lodash.isFunction(jqc.ready)) {
          const $$ = create$$(jqc.$fragment);
          const watcher = (key, val) => listenProps.on(jqc.id, key, val);
          $$.jqc = jqc;
          $$.on = watcher;
          $$.el = jqc.$fragment;
          jqc.ready.apply(jqc.$fragment, [proxy$$($$)]);
        }
      }
    }
    root.childNodes.forEach(child => {
      this.render(child);
    });
  }
}

var store = {};

function sortRouters(routers) {
  routers.sort((a, b) => a.path < b.path ? -1 : a.path > b.path ? 1 : 0).reverse();  for (let router of routers) {
    if (Array.isArray(router.children)) {
      sortRouters(router.children);
    }
  }
  return routers;
}

const setRouters = (routers) => {
  store.routers = sortRouters(routers);
};

const getRouters = () => {
  return store.routers;
};

function mountedComponet($, routers, base = '') {
  $.el.html('');

  let hash = location.hash.replace('#', '') || '/';

  const hasParentRouter = $.el.parent().closest(`[${COMPONENT_ROUTER}]`).attr(COMPONENT_ROUTER);

  for (let router of routers) {
    let { path, component, children } = router;
    let fullPath = `${base}${path}`;
    if (lodash.startsWith(hash, fullPath)) {
      if (hasParentRouter) {
        // 嵌套路由，继续找子路由
        if (lodash.isArray(children)) {
          return mountedComponet($, children, fullPath);
        }
      }
      if (lodash.isFunction(component)) {
        const componentPromise = component();
        componentPromise.then(comp => {
          $.el.attr(COMPONENT_ROUTER, fullPath);
          $.el.html(comp);
          listenProps.emit('all', 'router', { ...router, hash, path: fullPath });
        });
      } else {
        $.el.attr(COMPONENT_ROUTER, fullPath);
        $.el.html(component);
        listenProps.emit('all', 'router', { ...router, hash, path: fullPath });
      }
      return;
    }
  }
}

var router = {
  template: `<div ${COMPONENT_ROUTER}></div>`,
  ready($) {
    const listener = (evt) => routerChange();

    $.on('destroy', () => window.removeEventListener('hashchange', listener));
    window.addEventListener('hashchange', listener);

    routerChange();

    function routerChange() {
      let routers = getRouters();
      mountedComponet($, routers);
    }  }
};

const useJquery = jQuery => {

  function findInnerComponents($dom) {
    return $dom.find(`[${COMPONENT_ID}]`);
  }
  
  function broadcastDestroyComponents($dom) {
    const $domComponents = findInnerComponents($dom);
    $domComponents.each((index, dom) => {
      const componentId = jQuery(dom).attr(COMPONENT_ID);
      listenProps.emit(componentId, 'destroy', { componentId, dom });
    });
  }

  const $html = jQuery.prototype.html;
  const $attr = jQuery.prototype.attr;
  const $data = jQuery.prototype.data;

  jQuery.component = function (tagName, conf) {
    return setGlobalComponent(lodash.toUpper(tagName), conf);
  };

  jQuery.unComponent = function(tagName) {
    return removeGlobalComponent(lodash.toUpper(tagName))
  };

  jQuery.router = function (routers, tagName = 'router-view') {
    setRouters(routers);
    jQuery.component(tagName, router);
  };

  jQuery.prototype.attr = function () {
    let args = [...arguments];
    let [key, val] = args;
    let componentId = $attr.call(this, COMPONENT_ID);
    if (componentId) {
      listenProps.emit(componentId, key, val);
    }
    return $attr.apply(this, args);
  };

  jQuery.prototype.data = function () {
    let args = [...arguments];
    let [key, val] = args;
    let componentId = $attr.call(this, COMPONENT_ID);
    if (componentId) {
      listenProps.emit(componentId, key, val);
    }
    return $data.apply(this, args)
  };

  jQuery.prototype.html = function (obj) {
    // 通知销毁组件
    broadcastDestroyComponents(this);
    if (lodash.isObject(obj)) {
      // 在根节点下动态生成一个占位组件标签
      const templateApp = 'template-' + md5__default['default'](JSON.stringify(obj));
      $html.call(this, `<${templateApp}></${templateApp}>`);
      // 注册这个组件
      jQuery.component(templateApp, obj);
      // 则生成组件实例，再渲染
      let jqc = new JqComponent(obj);
      jqc.render(this.get(0));
      // 销毁动态组件
      jQuery.unComponent(templateApp);
      return this;
    } else {
      return $html.call(this, obj)
    }
  };
};

exports.useJquery = useJquery;
